#include "OcfEmbedded.h"

#ifdef OCFE_ENABLE_CHECKSUM
OcfeUInt16 OcfeFletcher16Checksum(const char *dataChar, OcfeUInt16 count)
{
	const unsigned char MaxWorkableLength = 20;
	OcfeUInt16 sum1 = 0xff, sum2 = 0xff;

	while (count) 
	{
		char workLength = count > MaxWorkableLength ? MaxWorkableLength : count;
		count -= workLength;
		do
		{
			sum2 += sum1 += *dataChar++;
		} while (--workLength);
		sum1 = (sum1 & 0xff) + (sum1 >> 8);
		sum2 = (sum2 & 0xff) + (sum2 >> 8);
	}
	sum1 = (sum1 & 0xff) + (sum1 >> 8);
	sum2 = (sum2 & 0xff) + (sum2 >> 8);
	return sum2 << 8 | sum1;
}
#endif

#ifdef OCFE_ENABLE_FRAME_NUMERATOR
#define OCFE_FRAME_POSITION_COMMAND				1
#define OCFE_FRAME_POSITION_FIRST_PARAMETER		2
#define OCFE_FRAME_POSITION_SECOND_PARAMETER	3
#else
#define OCFE_FRAME_POSITION_COMMAND				0
#define OCFE_FRAME_POSITION_FIRST_PARAMETER		1
#define OCFE_FRAME_POSITION_SECOND_PARAMETER	2
#endif

void FinalizeReturnValue(OcfeContext *context, char* valueBuffer)
{
	OcfeChar8 * OCFE_MEMORY_AREA buffer = context->Buffer;

	OCFE_STRING_COPY(&buffer[context->ElementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]], valueBuffer);
	context->PosEndString = context->ElementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER] + OCFE_STRING_LENGTH(valueBuffer) + 1;
#ifdef OCFE_ENABLE_CHECKSUM
	context->CheckSum = OcfeFletcher16Checksum(buffer, context->PosEndString);
	OCFE_SHORT_TO_STRING(&buffer[context->PosEndString], context->CheckSum);
	context->PosEndString += OCFE_STRING_LENGTH(&buffer[context->PosEndString]);
#endif
	buffer[context->PosEndString++] = OCFE_COMMAND_TERMINATOR;
}


OcfeUInt16 OcfeExecute(OcfeVariableEntry ocfeDictionnary[], OcfeContext *context, OcfeUInt16 inputDataLength)
{
	OCFE_MEMORY_AREA OcfeUInt8 i;
	OCFE_MEMORY_AREA OcfeUInt8 frameValid;
	OCFE_MEMORY_AREA OcfeUInt8 elementCount;
	OCFE_MEMORY_AREA OcfeInt16 index;
	OCFE_MEMORY_AREA OcfeChar8 numericBuffer[OCFE_MAX_VALUE_LENGTH];

	OcfeChar8 * OCFE_MEMORY_AREA buffer = context->Buffer;
	OcfeUInt8 * OCFE_MEMORY_AREA elementPositions = context->ElementPositions;
	OcfeUInt16 * OCFE_MEMORY_AREA posEndString = &context->PosEndString;
#ifdef OCFE_ENABLE_CHECKSUM
	OcfeUInt16 * OCFE_MEMORY_AREA checkSum = &context->CheckSum;
#endif
	
	index = 0;
	if (inputDataLength < sizeof(OcfeBuffer))
	{
		buffer[inputDataLength-1] = 0; // terminate frame with string terminal zero
		frameValid = 1;
		// localize elements
		elementPositions[0] = 0;
		elementCount = 1;
		for (i = 0; i < inputDataLength; i++)
			if (buffer[i] == OCFE_COMMAND_SEPARATOR)
			{
				if (elementCount < OCFE_MAX_FRAME_ELEMENT_COUNT)
				{
					elementPositions[elementCount] = i + 1;
					elementCount++;
				}
				else
				{
					frameValid = 0;
					break;
				}
			}
        
        // Process frame
		if (frameValid)
		{	
#ifdef OCFE_ENABLE_CHECKSUM
			// Check frame checksum
			OCFE_SHORT_FROM_STRING(*checkSum, &buffer[elementPositions[elementCount - 2]]);
			if ((*checkSum) == OcfeFletcher16Checksum(buffer, elementPositions[elementCount - 2]))
			{
#endif
                                    switch(buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]])
				{
				case OCFE_FRAME_TYPE_REQ_METADATA:
					for(i = 0; ocfeDictionnary[i].name != 0; i++)
						;
					OCFE_SHORT_FROM_STRING(index, &buffer[elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]]);
					if(index < i)
					{
						buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]] = OCFE_FRAME_TYPE_IND_METADATA;
						OCFE_STRING_COPY(&buffer[elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]], ocfeDictionnary[index].name);
						*posEndString = elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER] + OCFE_STRING_LENGTH(ocfeDictionnary[index].name)+1;
						OCFE_CHAR_TO_STRING(&buffer[(*posEndString)++], ocfeDictionnary[index].options);
					}
					else
					{
						buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]] = OCFE_FRAME_TYPE_IND_ERROR;
						OCFE_STRING_COPY(&buffer[elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]], OCFE_FRAME_INDEX_OUT_RANGE_MSG);
						*posEndString = elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER] + OCFE_FRAME_INDEX_OUT_RANGE_SIZE;
					}
#ifdef OCFE_ENABLE_CHECKSUM
					(*posEndString)++;
					*checkSum = OcfeFletcher16Checksum(buffer, *posEndString);
					OCFE_SHORT_TO_STRING(&buffer[*posEndString], *checkSum);
					*posEndString += OCFE_STRING_LENGTH(&buffer[*posEndString]);
#endif
					buffer[(*posEndString)++] = OCFE_COMMAND_TERMINATOR;
					return *posEndString;
				case OCFE_FRAME_TYPE_REQ_VARIABLE:
					while (ocfeDictionnary[index].name != 0)
					{
						if(OCFE_STRING_EQUAL(&buffer[elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]], ocfeDictionnary[index].name))
						{
							buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]] = OCFE_FRAME_TYPE_IND_VARIABLE;
							switch (ocfeDictionnary[index].type)
							{
#ifdef OCFE_ENABLE_BOOL
							case OCFE_TYPE_BOOL:
								LOCK
									OCFE_BOOL_TO_STRING(numericBuffer, *(char*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#endif
							case OCFE_TYPE_CHAR:
								LOCK
									OCFE_CHAR_TO_STRING(numericBuffer, *(char*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
							case OCFE_TYPE_SHORT:
								LOCK
									OCFE_SHORT_TO_STRING(numericBuffer, *(short*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#ifdef OCFE_ENABLE_INT
							case OCFE_TYPE_INT:
								LOCK
									OCFE_INT_TO_STRING(numericBuffer, *(int*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#endif
#ifdef OCFE_ENABLE_LONG
							case OCFE_TYPE_LONG:
								LOCK
									OCFE_LONG_TO_STRING(numericBuffer, *(long*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#endif
#ifdef OCFE_ENABLE_FLOAT
							case OCFE_TYPE_FLOAT:
								LOCK
									OCFE_FLOAT_TO_STRING(numericBuffer, *(float*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#endif
#ifdef OCFE_ENABLE_DOUBLE
							case OCFE_TYPE_DOUBLE:
								LOCK
									OCFE_DOUBLE_TO_STRING(numericBuffer, *(double*)ocfeDictionnary[index].address);
								UNLOCK
								FinalizeReturnValue(context, numericBuffer);
								break;
#endif
#ifdef OCFE_ENABLE_STRING
							case OCFE_TYPE_STRING:
								LOCK
									FinalizeReturnValue(context, (char*)ocfeDictionnary[index].address);
								UNLOCK
								break;
#endif
#ifdef OCFE_ENABLE_FUNCTION
							case OCFE_TYPE_FUNCTION_VOID:;
								(*(ptrFonctionVoid)ocfeDictionnary[index].address)();
								FinalizeReturnValue(context, (char*)"void");
								break;
							case OCFE_TYPE_FUNCTION_RET:
								FinalizeReturnValue(context, (*(pftFonctionRet)ocfeDictionnary[index].address)());
								break;
#endif
							default:
								break;
							}
							return *posEndString;
						}
						else
							index++;
					}
					frameValid = 0;
					break;
				case OCFE_FRAME_TYPE_REQ_WRITE:
					while (ocfeDictionnary[index].name != 0)
					{
						if(OCFE_STRING_EQUAL(&buffer[elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER]], ocfeDictionnary[index].name))
						{
							buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]] = OCFE_FRAME_TYPE_IND_WRITE;
							switch (ocfeDictionnary[index].type)
							{
#ifdef OCFE_ENABLE_BOOL
							case OCFE_TYPE_BOOL:
								LOCK
									OCFE_BOOL_FROM_STRING(*(char*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#endif
							case OCFE_TYPE_CHAR:
								LOCK
									OCFE_CHAR_FROM_STRING(*(char*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
							case OCFE_TYPE_SHORT:
								LOCK
									OCFE_SHORT_FROM_STRING(*(short*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#ifdef OCFE_ENABLE_INT
							case OCFE_TYPE_INT:
								LOCK
									OCFE_INT_FROM_STRING(*(int*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#endif
#ifdef OCFE_ENABLE_LONG
							case OCFE_TYPE_LONG:
								LOCK
									OCFE_LONG_FROM_STRING(*(long*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#endif
#ifdef OCFE_ENABLE_FLOAT
							case OCFE_TYPE_FLOAT:
								LOCK
									OCFE_FLOAT_FROM_STRING(*(float*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#endif
#ifdef OCFE_ENABLE_DOUBLE
							case OCFE_TYPE_DOUBLE:
								LOCK
									OCFE_DOUBLE_FROM_STRING(*(double*)ocfeDictionnary[index].address, &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
								UNLOCK
								break;
#endif
#ifdef OCFE_ENABLE_STRING
							case OCFE_TYPE_STRING:
								LOCK
									if(OCFE_STRING_LENGTH(&buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]) > ocfeDictionnary[index].size)
									{
										OCFE_STRING_COPY((char*)ocfeDictionnary[index].address,  &buffer[elementPositions[OCFE_FRAME_POSITION_SECOND_PARAMETER]]);
									}
									else
									{
										buffer[elementPositions[OCFE_FRAME_POSITION_COMMAND]] = OCFE_FRAME_TYPE_IND_ERROR;
										FinalizeReturnValue(context, (char*)OCFE_FRAME_ERROR_STRING_SIZE_OVERFLOW_MSG);
										return *posEndString;
									}
								UNLOCK
								break;
#endif
							default:
								break;
							}
							*posEndString = elementPositions[OCFE_FRAME_POSITION_FIRST_PARAMETER];
#ifdef OCFE_ENABLE_CHECKSUM
							*checkSum = OcfeFletcher16Checksum(buffer, *posEndString);
							OCFE_SHORT_TO_STRING(&buffer[*posEndString], *checkSum);
							*posEndString += OCFE_STRING_LENGTH(&buffer[*posEndString]);
#endif
							buffer[(*posEndString)++] = OCFE_COMMAND_TERMINATOR;

							return *posEndString;
						}
						else
							index++;
					}
					frameValid = 0;
					break;
				default:
					frameValid = 0;
					break;
				}

#ifdef OCFE_ENABLE_CHECKSUM
			}
			else
			{
				OCFE_STRING_COPY(buffer, OCFE_FRAME_CHECKSUM_ERROR_MSG);
				return OCFE_FRAME_CHECKSUM_ERROR_SIZE;
			}
#endif
		}
	}
	else
		frameValid = 0;

	if(!frameValid)
	{
		OCFE_STRING_COPY(buffer, OCFE_FRAME_TYPE_IND_INVALID);
		return OCFE_FRAME_TYPE_IND_INVALID_SIZE;
	}
	return 0;
}


OcfeUInt16 OcfeCharacterReceived(OcfeVariableEntry* ocfeDictionnary, OcfeStreamContext *context, char c)
{
	if (context->CharToWrite != 0 || context->Index > OCFE_COMMUNICATION_BUFFER_SIZE)
	{
		context->CharToWrite = 0;
		context->Index = 0;
	}

	context->context.Buffer[context->Index++] = c;
	if(c == OCFE_COMMAND_TERMINATOR)
	{
            context->CharToWrite = OcfeExecute(ocfeDictionnary, &(context->context), context->Index);
            context->Index = 0;
	}

	return context->CharToWrite;
}

OcfeUInt16 OcfeCharacterToSend(OcfeStreamContext *context, char* c)
{
	OCFE_MEMORY_AREA OcfeInt16 currentCharToWrite = context->CharToWrite - context->Index;

	if(currentCharToWrite > 0)
		*c = context->context.Buffer[context->Index++];
	else
		currentCharToWrite = 0;

	return currentCharToWrite;
}
